
abstract(Input,Output) :-
	see(Input), tell(Output), abstract_all,	seen, told.

abstract_all :- 
	read(Clause), 
	(Clause \== end_of_file -> 
	 abstract_clause(Clause), abstract_all
	  ;
	 true).

abstract_clause(( :- B)) :- !, nice_output(( :- B)).        % put it as it is
abstract_clause((H :- B)) :- !, abstract_atom(H,HA),        % usual clause
	abstract_body(B,BA), nice_output((HA :- BA)).
abstract_clause(H) :- abstract_atom(H,HA), nice_output(HA). % fact

abstract_body((B,Bs),(BA,BsA)) :- 
	abstract_atom(B,BA), !, abstract_body(Bs,BsA).
abstract_body(B,BA) :- abstract_atom(B,BA).

abstract_atom(A,AA) :- 
	A=..[N|Args], abstract_args(Args,AbstrArgs), AA=..[N|AbstrArgs].

abstract_args([],[]).
abstract_args([Arg|Args],[AbstrArg|AbstrArgs]) :- 
	abstr_term(Arg,AbstrArg), abstract_args(Args,AbstrArgs).

nice_output(Clause) :- numbervars(Clause,0,_), write_clause(Clause).

write_clause(( :- T)) :- write(' :- '), print(T), write('.'), nl,nl.
write_clause((H :- T)) :- print(H), write(' :- '), nl, write_body(T).
write_clause(H) :- print(H), write('.'), nl, nl.

write_body((T,Ts)) :- format('~8| ~p,~n',[T]), !, write_body(Ts).
write_body(T) :- format('~8| ~p.~n ~n',[T]).

%%-----------------------------------------------------------------------------

%% variable -- must be the first choice
abstr_term(V,[V]) :- var(V), !.

%% lists
abstr_term([],[nil]).
abstr_term([X|Xs],[list(TX),TXs]) :- abstr_term(X,TX), abstr_term(Xs,TXs).

%% trees
abstr_term(void,[void]).
abstr_term(tree(H,L,R),[tree(TH),TL,TR]) :-
	abstr_term(H,TH),
	abstr_term(L,TL), abstr_term(R,TR).

%% math expressions
abstr_term(MathExpr,[AbstrArg1,AbstrArg2]) :- 
	MathExpr=..[Op,Arg1,Arg2], 
	(Op==(+) ; Op==(*) ; Op==(-); Op==(/) ; Op==(^)), !,
	abstr_term(Arg1,AbstrArg1), abstr_term(Arg2,AbstrArg2).

abstr_term(MathExpr,[AbstrArg]) :- 
	MathExpr=..[Func,Arg],
	(
	    Func==(-) ; Func==sin ; Func==cos ; Func==abs ;
	    Func==arcsin ; Func==arccos
	), !,
	abstr_term(Arg,AbstrArg).

%% other simple types. Types like "char", "string" etc. can be added here
abstr_term(Num,[num]) :- number(Num), !.
abstr_term(Atom,[atom]) :- atom(Atom), !.

 

abstr_term(_,[other]).				% always the last choice

/*
abstr_term(_,[_]).	% alternative -- abstracting all "other" types
			% to "top" (a new variable)
*/
