
:- module(aci_unify,[aci_unify/2, unify_atoms/2]).
:- use_module(aci_basics).
:- ensure_loaded([library(lists), library(ordsets)]).

unify_atoms([N|Args1],[N|Args2]) :- aci_unify(Args1,Args2).

aci_unify([],[]).
aci_unify([A|As],[B|Bs]) :-
	flatten(A,AF), flatten(B,BF),
	unify_one_pair(AF,BF),
	aci_unify(As,Bs).

unify_one_pair(A,B) :- ground(A), ground(B), !, A==B.
unify_one_pair(A,B) :-
	ord_union(A,B,ToCover), ord_setproduct(A,B,AxB),
	covering_u_table(AxB,ToCover,CovUTable),
	keysort(CovUTable,CovUTableS),
	table2graph(CovUTableS,UG),
	u_graph_proceed(UG).

select_ground([],[]).
select_ground([X|Xs],O) :- (ground(X) -> O = [X|GXs] ; O = GXs),
	select_ground(Xs,GXs).

covering_u_table([],[],[]).
covering_u_table([X-Y|XsYs],ToCover,[X-L,Y-L|UTRest]) :-
	(X==Y -> ord_subtract(ToCover,[X],ToCoverRest), !
	  ;
	 (X @< Y ->
	  ord_subtract(ToCover,[X,Y],ToCoverRest)
	   ;
	  ord_subtract(ToCover,[Y,X],ToCoverRest))),
	 covering_u_table(XsYs,ToCoverRest,UTRest).
covering_u_table([_|XsYs],ToCover,UTable) :-
	covering_u_table(XsYs,ToCover,UTable).

table2graph([],[]).
table2graph([V-I],[V-[I]]) :- !.
table2graph([V-I,W-J|Vs],Out) :-
	(V==W ->
	  Out = [V-[I|Is]|VsClps], Cont = [V-Is|VsClps]
	  ;
	  Out = [V-[I]|VsClps], Cont = VsClps),
	table2graph([W-J|Vs],Cont).

u_graph_proceed([]).
u_graph_proceed([V-E|G]) :- unify_single(V,E), u_graph_proceed(G).

unify_single(V,Y) :- var(V), !, V=Y.		    % variable
unify_single(S,Y) :- simple(S), !, match_list(Y,S). % simple (not f(...))
unify_single(T,Y) :- T=..[N,Arg],		    % only unary functions!
	lower_level(Y,N,YArgs),
	dup_n(YArgs,Arg,Args),
	aci_unify(YArgs,Args).

lower_level([],_,[]).
lower_level([NT|NTs],N,[T|Ts]) :- NT=..[N,T], lower_level(NTs,N,Ts).

dup_n([],_,[]).
dup_n([_|Xs],Y,[Y|Ys]) :- dup_n(Xs,Y,Ys).

match_list([],_).
match_list([X|Xs],X) :- match_list(Xs,X).

