
:- module(canonical,[canonical/2]).
:- use_module(aci_basics).
:- ensure_loaded([library(ordsets)]). %, library(basics)]).

canonical([N|Args],[N|ArgsC]) :-
	flatten_seq(Args,ArgsF),
	occur_graph(ArgsF,OG),
	swap_keys_values(OG,GO),
	keysort(GO,GOS),
	drop_equiv(GOS,NewGO),
	graph2table(NewGO,OGTable),
	keysort(OGTable,OGTableS),
	table2graph(OGTableS,KeysArgs),
	keys_and_values(KeysArgs,_,ArgsC).

drop_equiv([],[]).
drop_equiv([X],[X]) :- !.
drop_equiv([E1-V1,E2-V2|Vs],Out) :-
	(E1==E2, variant(V1,V2) ->
	 V1=V2, Out=OutRest
	  ;
	 Out=[E1-V1|OutRest]),
	drop_equiv([E2-V2|Vs],OutRest).

occur_graph(ArgsVs,OG) :-
	og_table(ArgsVs,0,OG_Table),
	sort(OG_Table,OG_TableSorted),
	table2graph(OG_TableSorted,OG).

og_table([],_,[]).
og_table([[]|Args],I,OG_Table) :- !, J is I+1, og_table(Args,J,OG_Table).
og_table([[V|Vs]|Args],I,[V-I|OG_Table]) :- og_table([Vs|Args],I,OG_Table).

table2graph([],[]).
table2graph([V-I],[V-[I]]) :- !.
table2graph([V-I,W-J|Vs],Out) :-
	(V==W ->
	 Out = [V-[I|Is]|VsClps], Cont = [V-Is|VsClps]
	  ;
	 Out = [V-[I]|VsClps], Cont = VsClps),
	table2graph([W-J|Vs],Cont).

graph2table([],[]).
graph2table([[]-_|Vs],Edges) :- !, graph2table(Vs,Edges).
graph2table([[E|Es]-V|Vs],[E-V|Edges]) :- graph2table([Es-V|Vs],Edges).

swap_keys_values([],[]).
swap_keys_values([Key-Val|KVs],[Val-Key|VKs]) :- swap_keys_values(KVs,VKs).
