
:- ensure_loaded(library(lists)).
:- use_module([aci_unify, aci_basics, canonical, builtins]).
:- dynamic(fact/2).
:- dynamic(database_changed/0).

go :- go(user).

go(Out) :-
	retractall(fact(_,_)),
	statistics(runtime,_),			% init runtime statistics
	assert_facts,
	iterate_tp(1),
	statistics(runtime,[_,T]),
	tell(Out), showfacts, told,		% model output
	RTime is T / 1000,
	format('~n-----~n~3f sec. runtime~n',[RTime]).

assert_facts :- my_clause(H,[]), assert(fact(H,0)), fail.
assert_facts.

%%-----------------------------------------------------------------------------

iterate_tp(N) :- N1 is N-1,
	fact(Atom,N1), new(Atom,NewF),
	cond_assert(NewF,N), fail.
iterate_tp(N) :- retract(database_changed), N1 is N+1, iterate_tp(N1).
iterate_tp(_).
	
new(Fact,H) :- pattern(Fact,FP),
	my_clause(H,B), select(FP,B,Rest),
	unify_atoms(Fact,FP), solve(Rest).

solve([]).
solve([X|Xs]) :- pattern(X,XP),
	(builtin(XP), ! ; fact(XP,_)),
	unify_atoms(X,XP),
	solve(Xs).

%%-----------------------------------------------------------------------------
cond_assert(F,N) :- canonical(F,FC), %print(FC), nl,
        \+ (numbervars(FC,0,_), fact(FC,_)), assertz(fact(FC,N)),
	(clause(database_changed,true) -> true ; assert(database_changed)).

pattern([N|Args],[N|ArgsP]) :- same_length(Args,ArgsP).

%%-----------------------------------------------------------------------------
%% Pretty print
showfacts :-
	fact([N|Args],_),
	numbervars(Args,0,_),
	A=..[N|Args],
	print(A), nl,
	fail ; true.

