
:- ensure_loaded(library(lists)).
:- use_module([aci_unify, aci_basics, canonical, builtins]).
:- dynamic(fact/2).
:- dynamic(database_changed/0).

go(InitCall) :- go(InitCall,user).

go(InitCall,Out) :-
	retractall(fact(_,_)),
	statistics(runtime,_),			% init runtime statistics
	cond_assert(InitCall-call,0),
	iterate_tp(1),
	statistics(runtime,[_,T]),
	tell(Out), showfacts, told,		% model output
	RTime is T / 1000,
	format('~n-----~n~3f sec. runtime~n',[RTime]).

%%-----------------------------------------------------------------------------

iterate_tp(N) :- N1 is N-1,
	fact(Atom,N1), new(Atom,NewF),
	cond_assert(NewF,N), fail.
iterate_tp(N) :- retract(database_changed), !, N1 is N+1, iterate_tp(N1).
iterate_tp(_).

new(Fact-call,NewF) :-
	pattern(Fact,FP), my_clause(FP,B), unify_atoms(FP,Fact),
	solve_right(B,Fact,NewF).

new(Fact-ans,NewF) :-
	my_clause(H,B), inbody(Fact,B,Left,Right),
	pattern(H,HP), fact(HP-call,_), unify_atoms(H,HP),
	solve_left(Left), solve_right(Right,H,NewF).

inbody(Fact,[B|Bs], [], Bs) :- unify_atoms(Fact,B).
inbody(Fact,[B|Bs], [B|Left], Right) :- inbody(Fact,Bs,Left,Right).

solve_left([]).
solve_left([X|Xs]) :- pattern(X,XP),
	(builtin(XP), ! ; fact(XP-ans,_)), unify_atoms(X,XP),
	solve_left(Xs).

solve_right([],Head,Head-ans).
solve_right([X|Xs],Head,NewF) :- pattern(X,XP),
	(builtin(XP) ->
	 unify_atoms(X,XP), solve_right(Xs,Head,NewF)
	  ;
	 (NewF = X-call
	   ;
	  fact(XP-ans,_), unify_atoms(X,XP), solve_right(Xs,Head,NewF)
	 )).

%%-----------------------------------------------------------------------------
cond_assert(F-Tag,N) :- canonical(F,FC),
        \+ (numbervars(FC,0,_), fact(FC-Tag,_)), assertz(fact(FC-Tag,N)),
	(clause(database_changed,true) -> true ; assert(database_changed)).

pattern([N|Args],[N|ArgsP]) :- same_length(Args,ArgsP).

%%-----------------------------------------------------------------------------
%% Pretty print
showfacts :-
	fact([N|Args]-Tag,_),
	numbervars(Args,0,_),
	A=..[N|Args],
	print(A-Tag), nl,
	fail ; true.
